/*
 * USB to SPI/I2C/GPIO master driver for USB converter chip ch347/ch341, etc.
 *
 * Copyright (C) 2023 Nanjing Qinheng Microelectronics Co., Ltd.
 * Web: http://wch.cn
 * Author: WCH <tech@wch.cn>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Update Log:
 * V1.0 - initial version
 * V1.1 - add supports for i2c master, gpio irq function
 */

#include "ch34x_mpsi.h"

static DEFINE_IDA(ch34x_devid_ida);

/* table of devices that work with this driver */
static const struct usb_device_id ch34x_usb_ids[] = {
	{ USB_DEVICE(0x1a86, 0x5512) },
	{ USB_DEVICE_INTERFACE_NUMBER(0x1a86, 0x55db, 0x02) }, /* CH347 Mode1 SPI+IIC+UART */
	{}						       /* Terminating entry */
};

MODULE_DEVICE_TABLE(usb, ch34x_usb_ids);

struct ch341_pin_config ch341_board_config[CH341_CS_NUM] = {
	{ 15, "cs0" },
	{ 16, "cs1" },
	{ 17, "cs2" },
};

struct ch347_pin_config ch347_board_config[CH347_MPSI_GPIOS] = {
	{ 15, "gpio4", 4, GPIO_MODE_OUT, false },
	{ 2, "gpio6", 6, GPIO_MODE_IN, false },
	{ 13, "gpio7", 7, GPIO_MODE_OUT, false },
};

extern struct spi_board_info ch341_spi_device_template;
extern struct spi_board_info ch347_spi_device_template;
extern struct spi_board_info ch34x_spi_devices[CH341_SPI_MAX_NUM_DEVICES];

extern int ch34x_mpsi_spi_probe(struct ch34x_device *ch34x_dev);
extern int ch34x_mpsi_spi_remove(struct ch34x_device *ch34x_dev);
extern int ch34x_spi_probe(struct ch34x_device *ch34x_dev);
extern void ch34x_spi_remove(struct ch34x_device *ch34x_dev);
extern int ch34x_mpsi_i2c_probe(struct ch34x_device *ch34x_dev);
extern void ch34x_mpsi_i2c_remove(struct ch34x_device *ch34x_dev);
extern int ch347_irq_probe(struct ch34x_device *ch34x_dev);
extern void ch347_irq_remove(struct ch34x_device *ch34x_dev);
extern int ch34x_mpsi_gpio_probe(struct ch34x_device *ch34x_dev);
extern void ch34x_mpsi_gpio_remove(struct ch34x_device *ch34x_dev);
extern int ch347_irq_check(struct ch34x_device *ch34x_dev, u8 irq);

static int ch34x_cfg_probe(struct ch34x_device *ch34x_dev)
{
	struct ch341_pin_config *ch341cfg;
	struct ch347_pin_config *ch347cfg;
	int i;

	CHECK_PARAM_RET(ch34x_dev, -EINVAL);

	if (ch34x_dev->chiptype == CHIP_CH341) {
		/* set out: mosi/out2/sck/cs, in: miso */
		ch34x_dev->gpio_mask = 0x3f;
		for (i = 0; i < CH341_CS_NUM; i++) {
			ch341cfg = ch341_board_config + i;

			ch34x_spi_devices[ch34x_dev->slave_num] = ch341_spi_device_template;
			ch34x_spi_devices[ch34x_dev->slave_num].bus_num = CH341_SPI_BUS_NUM;
			ch34x_spi_devices[ch34x_dev->slave_num].mode = CH341_SPI_MODE;
			ch34x_spi_devices[ch34x_dev->slave_num].chip_select = ch341cfg->pin - 15;

			DEV_INFO(CH34X_USBDEV, "output %s SPI slave with CS%d", ch341cfg->name,
				 ch34x_spi_devices[ch34x_dev->slave_num].chip_select);

			ch34x_dev->slave_num++;
		}
	} else {
		ch34x_dev->irq_num = 0;
		ch34x_dev->irq_base = 0;
		// ch34x_dev->irq_hw = -1;

		for (i = 0; i < CH347_CS_NUM; i++) {
			ch34x_spi_devices[ch34x_dev->slave_num] = ch347_spi_device_template;
			ch34x_spi_devices[ch34x_dev->slave_num].bus_num = CH347_SPI_BUS_NUM;
			ch34x_spi_devices[ch34x_dev->slave_num].mode = CH347_SPI_MODE;
			ch34x_spi_devices[ch34x_dev->slave_num].chip_select = i;

			DEV_INFO(CH34X_USBDEV, "output SPI slave with CS%d",
				 ch34x_spi_devices[ch34x_dev->slave_num].chip_select);

			ch34x_dev->slave_num++;
		}
		for (i = 0; i < CH347_MPSI_GPIOS; i++) {
			ch347cfg = ch347_board_config + i;
			ch34x_dev->gpio_names[ch34x_dev->gpio_num] = ch347cfg->name;
			ch34x_dev->gpio_pins[ch34x_dev->gpio_num] = ch347cfg;
			ch34x_dev->gpio_irq_map[ch34x_dev->gpio_num] = ch34x_dev->irq_num;
			ch34x_dev->irq_gpio_map[ch34x_dev->irq_num] = ch34x_dev->gpio_num;

			// if (ch347cfg->hwirq) {
			// 	if (ch34x_dev->irq_hw != -1) {
			// 		DEV_ERR(CH34X_USBDEV,
			// 			"pin %d: only one GPIO can be connected to the hardware IRQ",
			// 			ch347cfg->pin);
			// 		return -EINVAL;
			// 	}
			// 	ch34x_dev->irq_hw = ch34x_dev->irq_num;
			// }

			DEV_INFO(CH34X_USBDEV, "%s %s irq=%d %s", ch347cfg->mode == GPIO_MODE_IN ? "input " : "output",
				 ch347cfg->name, ch34x_dev->irq_num, ch347cfg->hwirq ? "(hwirq)" : "");

			ch34x_dev->irq_num++;
			ch34x_dev->gpio_num++;
		}
	}
	return 0;
}

static void ch34x_cfg_remove(struct ch34x_device *ch34x_dev)
{
	CHECK_PARAM(ch34x_dev);

	return;
}

int ch34x_usb_transfer(struct ch34x_device *ch34x_dev, int out_len, int in_len)
{
	int retval;
	int actual = 0;
	int size;

	if (out_len != 0) {
		retval = usb_bulk_msg(ch34x_dev->usb_dev,
				      usb_sndbulkpipe(ch34x_dev->usb_dev, usb_endpoint_num(ch34x_dev->bulk_out)),
				      ch34x_dev->bulkout_buf, out_len, &actual, 2000);
		if (retval < 0)
			return retval;
	}

	if (in_len == 0)
		return actual;

	if (ch34x_dev->chiptype == CHIP_CH341)
		size = CH341_USB_MAX_BULK_SIZE;
	else
		size = in_len;

	memset(ch34x_dev->bulkin_buf, 0, MAX_BUFFER_LENGTH * 2);
	retval = usb_bulk_msg(ch34x_dev->usb_dev,
			      usb_rcvbulkpipe(ch34x_dev->usb_dev, usb_endpoint_num(ch34x_dev->bulk_in)),
			      ch34x_dev->bulkin_buf, size, &actual, 2000);

	if (retval < 0)
		return retval;

	return actual;
}

static void ch34x_usb_complete_intr_urb(struct urb *urb)
{
	struct ch34x_device *ch34x_dev;
	u8 gpioindex;
	bool triggered;
	int i;

	CHECK_PARAM(urb);
	CHECK_PARAM(ch34x_dev = urb->context);
  
	if (!urb->status) {
		DEV_DBG(CH34X_USBDEV, "%d", urb->status);

		for (i = 0; i < CH347_MPSI_GPIOS; i++) {
			gpioindex = ch34x_dev->gpio_pins[i]->gpioindex;
			triggered = ch34x_dev->intrin_buf[gpioindex+3] & BIT(3);
			DEV_DBG(CH34X_USBDEV, "enable[%d]=%d gpioindex=%d triggered=%d", i, ch34x_dev->irq_enabled[i],
				gpioindex, triggered);
			if (ch34x_dev->irq_enabled[i] && triggered)
				ch347_irq_check(ch34x_dev, i);
		}

		usb_submit_urb(ch34x_dev->intr_urb, GFP_ATOMIC);
	}
}

static void ch34x_usb_free_device(struct ch34x_device *ch34x_dev)
{
	CHECK_PARAM(ch34x_dev)

	if (ch34x_dev->intr_urb)
		usb_free_urb(ch34x_dev->intr_urb);

	usb_set_intfdata(ch34x_dev->intf, NULL);
	usb_put_dev(ch34x_dev->usb_dev);

	kfree(ch34x_dev);
}

static int ch34x_usb_probe(struct usb_interface *intf, const struct usb_device_id *id)
{
	struct usb_device *usb_dev = usb_get_dev(interface_to_usbdev(intf));
	struct usb_endpoint_descriptor *endpoint;
	struct usb_host_interface *iface_desc;
	struct ch34x_device *ch34x_dev;
	int i;
	int ret;
	int length;

	DEV_DBG(&intf->dev, "connect device");

	ch34x_dev = kzalloc(sizeof(struct ch34x_device), GFP_KERNEL);
	if (!ch34x_dev) {
		DEV_ERR(&intf->dev, "could not allocate device memor");
		return -ENOMEM;
	}

	ch34x_dev->usb_dev = usb_dev;
	ch34x_dev->intf = intf;
	iface_desc = intf->cur_altsetting;

	DEV_DBG(CH34X_USBDEV, "bNumEndpoints=%d", iface_desc->desc.bNumEndpoints);

	for (i = 0; i < iface_desc->desc.bNumEndpoints; i++) {
		endpoint = &iface_desc->endpoint[i].desc;

		DEV_DBG(CH34X_USBDEV, "  endpoint=%d type=%d dir=%d addr=%0x", i, usb_endpoint_type(endpoint),
			usb_endpoint_dir_in(endpoint), usb_endpoint_num(endpoint));

		if (usb_endpoint_is_bulk_in(endpoint)) {
			ch34x_dev->bulk_in = endpoint;
			ch34x_dev->bulkin_buf = kmalloc(MAX_BUFFER_LENGTH * 2, GFP_KERNEL);
		}

		else if (usb_endpoint_is_bulk_out(endpoint)) {
			ch34x_dev->bulk_out = endpoint;
			ch34x_dev->bulkout_buf = kmalloc(MAX_BUFFER_LENGTH * 2, GFP_KERNEL);
		}

		else if (usb_endpoint_xfer_int(endpoint)) {
			ch34x_dev->intr_in = endpoint;
			ch34x_dev->intrin_buf = kmalloc(CH347_USB_MAX_INTR_SIZE, GFP_KERNEL);
		}
	}

	if (id->idProduct == 0x5512) {
		ch34x_dev->chiptype = CHIP_CH341;
		length = CH341_USB_MAX_INTR_SIZE;
	} else {
		ch34x_dev->chiptype = CHIP_CH347;
		length = CH347_USB_MAX_INTR_SIZE;
	}

	/* save the pointer to the new ch34x_device in USB interface device data */
	usb_set_intfdata(intf, ch34x_dev);
	mutex_init(&ch34x_dev->io_mutex);

	ch34x_dev->id = ida_simple_get(&ch34x_devid_ida, 0, 0, GFP_KERNEL);
	if (ch34x_dev->id < 0)
		return ch34x_dev->id;

	ret = ch34x_cfg_probe(ch34x_dev);
	if (ret < 0)
		goto error;

	ret = ch34x_mpsi_spi_probe(ch34x_dev);
	if (ret < 0)
		goto error;

	ret = ch34x_spi_probe(ch34x_dev);
	if (ret < 0)
		goto error;

	ret = ch34x_mpsi_i2c_probe(ch34x_dev);
	if (ret < 0)
		goto error;

	ret = ch347_irq_probe(ch34x_dev);
	if (ret < 0)
		goto error;

	if (ch34x_dev->chiptype == CHIP_CH347) {
		ret = ch34x_mpsi_gpio_probe(ch34x_dev);
		if (ret < 0)
			goto error;
	}

	if (ch34x_dev->intr_in) {
		/* create URBs for handling interrupts */
		if (!(ch34x_dev->intr_urb = usb_alloc_urb(0, GFP_KERNEL))) {
			DEV_ERR(&intf->dev, "failed to alloc URB");
			ret = -ENOMEM;
			goto error;
		}

		usb_fill_int_urb(ch34x_dev->intr_urb, ch34x_dev->usb_dev,
				 usb_rcvintpipe(ch34x_dev->usb_dev, usb_endpoint_num(ch34x_dev->intr_in)),
				 ch34x_dev->intrin_buf, length, ch34x_usb_complete_intr_urb, ch34x_dev,
				 ch34x_dev->intr_in->bInterval);

		usb_submit_urb(ch34x_dev->intr_urb, GFP_ATOMIC);
	}

	DEV_INFO(CH34X_USBDEV, "USB to SPI/I2C/GPIO adapter ch34x now attached.");

	return 0;

error:
	ch34x_mpsi_i2c_remove(ch34x_dev);
	ch34x_spi_remove(ch34x_dev);
	ch34x_mpsi_spi_remove(ch34x_dev);
	ida_simple_remove(&ch34x_devid_ida, ch34x_dev->id);
	if (ch34x_dev->bulkin_buf)
		kfree(ch34x_dev->bulkin_buf);
	if (ch34x_dev->bulkout_buf)
		kfree(ch34x_dev->bulkout_buf);
	if (ch34x_dev->intrin_buf)
		kfree(ch34x_dev->intrin_buf);
	ch34x_usb_free_device(ch34x_dev);

	return ret;
}

static void ch34x_usb_disconnect(struct usb_interface *intf)
{
	struct ch34x_device *ch34x_dev = usb_get_intfdata(intf);

	DEV_INFO(CH34X_USBDEV, "ch34x adapter now disconnected");

	ch34x_mpsi_gpio_remove(ch34x_dev);
	ch347_irq_remove(ch34x_dev);
	ch34x_mpsi_i2c_remove(ch34x_dev);
	ch34x_spi_remove(ch34x_dev);
	ch34x_mpsi_spi_remove(ch34x_dev);
	ch34x_cfg_remove(ch34x_dev);
	ida_simple_remove(&ch34x_devid_ida, ch34x_dev->id);
	if (ch34x_dev->bulkin_buf)
		kfree(ch34x_dev->bulkin_buf);
	if (ch34x_dev->bulkout_buf)
		kfree(ch34x_dev->bulkout_buf);
	if (ch34x_dev->intrin_buf)
		kfree(ch34x_dev->intrin_buf);
	ch34x_usb_free_device(ch34x_dev);
}

static struct usb_driver ch34x_usb_driver = { .name = "mpsi-ch34x",
					      .id_table = ch34x_usb_ids,
					      .probe = ch34x_usb_probe,
					      .disconnect = ch34x_usb_disconnect };

module_usb_driver(ch34x_usb_driver);

MODULE_ALIAS(DRIVER_ALIAS);
MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
